import bpy
from mathutils import Euler
from bpy.props import *
from bpy.types import Operator
from bpy_extras import anim_utils
import gpu
from gpu_extras.batch import batch_for_shader


def smart_bake_pose(obj, use_existing=False, frame_start=1, frame_end=250, **kwargs):
    act = getattr(obj.animation_data, "action", None)
    if act is None:
        return

    keyed_frames = set()

    for curve in act.fcurves:
        frame_nums = [0.0 for i in range(len(curve.keyframe_points) * 2)]
        curve.keyframe_points.foreach_get("co", frame_nums)
        for fnum in frame_nums[::2]:
            keyed_frames.add(int(fnum))

    frames = sorted(filter(lambda x: frame_start <=
                    x <= frame_end, keyed_frames))
    anim_utils.bake_action(
        obj, action=act if use_existing else None, frames=frames, **kwargs)


class NLA_OT_smart_bake(Operator):
    """Bake all selected objects loc/scale/rotation animation to an action"""
    bl_idname = "nla.bake"
    bl_label = "Bake Action"
    bl_options = {'REGISTER', 'UNDO'}

    frame_start: IntProperty(
        name="Start Frame",
        description="Start frame for baking",
        min=0, max=300000,
        default=1,
    )
    frame_end: IntProperty(
        name="End Frame",
        description="End frame for baking",
        min=1, max=300000,
        default=250,
    )
    step: IntProperty(
        name="Frame Step",
        description="Frame Step",
        min=1, max=120,
        default=1,
    )
    only_selected: BoolProperty(
        name="Only Selected Bones",
        description="Only key selected bones (Pose baking only)",
        default=True,
    )
    visual_keying: BoolProperty(
        name="Visual Keying",
        description="Keyframe from the final transformations (with constraints applied)",
        default=False,
    )
    clear_constraints: BoolProperty(
        name="Clear Constraints",
        description="Remove all constraints from keyed object/bones, and do 'visual' keying",
        default=False,
    )
    clear_parents: BoolProperty(
        name="Clear Parents",
        description="Bake animation onto the object then clear parents (objects only)",
        default=False,
    )
    use_current_action: BoolProperty(
        name="Overwrite Current Action",
        description="Bake animation into current action, instead of creating a new one "
                    "(useful for baking only part of bones in an armature)",
        default=False,
    )
    clean_curves: BoolProperty(
        name="Clean Curves",
        description="After baking curves, remove redundant keys",
        default=False,
    )
    bake_types: EnumProperty(
        name="Bake Data",
        description="Which data's transformations to bake",
        options={'ENUM_FLAG'},
        items=(
            ('POSE', "Pose", "Bake bones transformations"),
            ('OBJECT', "Object", "Bake object transformations"),
        ),
        default={'POSE'},
    )
    do_smart_bake: BoolProperty(
        name="Use Smart Bake",
        description="Bake pose animation only on existing keyframes",
        default=False
    )

    def execute(self, context):
        from bpy_extras import anim_utils
        do_pose = 'POSE' in self.bake_types
        do_object = 'OBJECT' in self.bake_types

        objects = context.selected_editable_objects

        # If we have Smart Bake enabled, remove all pose objects and smart bake them separately
        if self.do_smart_bake and 'POSE' in self.bake_types:
            non_pose_obs = []
            for obj in objects:
                if obj.type == 'ARMATURE':
                    smart_bake_pose(obj, use_existing=self.use_current_action, frame_start=self.frame_start,
                                    frame_end=self.frame_end,
                                    only_selected=self.only_selected,
                                    do_visual_keying=self.visual_keying,
                                    do_constraint_clear=self.clear_constraints,
                                    do_parents_clear=self.clear_parents,
                                    do_clean=True, do_pose=True, do_object=False)
                else:
                    non_pose_obs.append(obj)
            objects = non_pose_obs

        # If we have Smart Bake enabled, remove all object objects and smart bake them separately
        if self.do_smart_bake and 'OBJECT' in self.bake_types:
            for obj in objects:
                smart_bake_pose(obj, use_existing=self.use_current_action, frame_start=self.frame_start,
                                frame_end=self.frame_end,
                                only_selected=self.only_selected,
                                do_visual_keying=self.visual_keying,
                                do_constraint_clear=self.clear_constraints,
                                do_parents_clear=self.clear_parents,
                                do_clean=True, do_pose=False, do_object=True)
            return {'FINISHED'}

        if not len(objects):
            return {'FINISHED'}

        object_action_pairs = (
            [(obj, getattr(obj.animation_data, "action", None))
             for obj in objects]
            if self.use_current_action else
            [(obj, None) for obj in objects]
        )

        actions = anim_utils.bake_action_objects(
            object_action_pairs,
            frames=range(self.frame_start, self.frame_end + 1, self.step),
            only_selected=self.only_selected,
            do_pose=do_pose,
            do_object=do_object,
            do_visual_keying=self.visual_keying,
            do_constraint_clear=self.clear_constraints,
            do_parents_clear=self.clear_parents,
            do_clean=self.clean_curves,
        )

        if not any(actions):
            self.report({'INFO'}, "Nothing to bake")
            return {'CANCELLED'}

        return {'FINISHED'}

    def invoke(self, context, _event):
        scene = context.scene
        self.frame_start = scene.frame_start
        self.frame_end = scene.frame_end
        self.bake_types = {'POSE'} if context.mode == 'POSE' else {'OBJECT'}

        wm = context.window_manager
        return wm.invoke_props_dialog(self)


constraint_types = [
    ('NONE', "None", "Don't add constraints to the objects/bones after baking"),
    ('COPY_LOCATION', "Copy Location",
     "Add Copy Location constraints to affected objects/bones"),
    ('COPY_ROTATION', "Copy Rotation",
     "Add Copy Rotation constraints to affected objects/bones"),
    ('COPY_TRANSFORMS', "Copy Transforms",
     "Add Copy Transforms constraints to affected objects/bones"),
    ('DAMPED_TRACK', "Damped Track",
     "Add Damped Track constraints to affected objects/bones")
]


def draw_callback_px(self, context):

    gpu.state.blend_set('ALPHA')
    gpu.state.line_width_set(2.0)

    # 50% alpha, 2 pixel width line
    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
    batch = batch_for_shader(shader, 'POINTS', {"pos": self.emp_loc})
    shader.uniform_float("color", (1.0, 1.0, 1.0, 1.0))
    batch.draw(shader)

    # 50% alpha, 2 pixel width line
    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
    batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": self.x_line})
    shader.uniform_float("color", (1.0, 0.0, 0.0, 0.5))
    batch.draw(shader)

    # 50% alpha, 2 pixel width line
    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
    batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": self.y_line})
    shader.uniform_float("color", (0.0, 1.0, 0.0, 0.5))
    batch.draw(shader)

    # 50% alpha, 2 pixel width line
    shader = gpu.shader.from_builtin('UNIFORM_COLOR')
    batch = batch_for_shader(shader, 'LINE_STRIP', {"pos": self.z_line})
    shader.uniform_float("color", (0.0, 0.0, 1.0, 0.5))
    batch.draw(shader)

    # restore opengl defaults
    gpu.state.line_width_set(1.0)
    gpu.state.blend_set('NONE')


def is_a_library(data_path):
    is_library = data_path.library is not None
    is_lo = data_path.override_library is not None and data_path.override_library.reference.library is not None

    return is_library or is_lo


class ARM_OT_BakeBonesToEmpties(Operator):
    bl_idname = 'object.bake_to_empties'
    bl_label = 'Bake to empties'
    bl_options = {'UNDO'}

    frame_start: IntProperty(name="Start", default=1)
    frame_end: IntProperty(name="End", default=100)

    bake: BoolProperty(name="Bake animation", default=True)
    use_scene: BoolProperty(name="Use scene start/end", default=True)
    use_custom_space: BoolProperty(name="Use Custom Space", default=False)
    use_custom_axis: BoolProperty(name="Use Custom Axis", default=False)
    use_additive: BoolProperty(name="Use Additive Layer", default=False)

    subcol_name: StringProperty(
        name="Collection Name",
        description="Collection to dump baked empties into",
        default="Baked Empties"
    )

    bone_prefix: StringProperty(
        name="Prefix",
        description="Prefix for empty names (e.g. bake_Bone.001)",
        default="bake_"
    )

    custom_space: StringProperty(
        name="Custom Space",
        description="Parent empties to this object and bake."
    )

    custom_space_bone: StringProperty(
        name="Bone",
        description="Parent empties to this bone and bake."
    )

    constraint: EnumProperty(
        items=constraint_types,
        name="Constraint",
        description="Constraint type to add to bones",
        default='COPY_TRANSFORMS'
    )

    @classmethod
    def poll(cls, context):
        return context.object is not None and context.mode in {'OBJECT', 'POSE'}

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type == 'MOUSEMOVE':

            fac = 0.005
            if event.shift:
                fac *= 0.1

            self.length -= (self.start_x - event.mouse_x) * fac

            if self.axis == 'X':
                axis = context.active_pose_bone.x_axis
            elif self.axis == 'Y':
                axis = context.active_pose_bone.y_axis
            elif self.axis == 'Z':
                axis = context.active_pose_bone.z_axis

            head = context.active_pose_bone.head
            vec = axis * context.active_pose_bone.bone.length

            self.emp_loc.clear()
            self.emp_loc.append(head + vec * self.length)

            bpy.context.window.cursor_warp(self.start_x, self.start_y)

        elif event.type == 'X' and event.value == 'RELEASE':
            self.axis = 'X'
        elif event.type == 'Y' and event.value == 'RELEASE':
            self.axis = 'Y'
        elif event.type == 'Z' and event.value == 'RELEASE':
            self.axis = 'Z'

        elif event.type == 'LEFTMOUSE':
            bpy.context.window.cursor_modal_set('DEFAULT')
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')

            col = bpy.data.collections.new(self.subcol_name)
            parent = context.collection

            if parent and not is_a_library(parent):
                context.collection.children.link(col)
            else:
                context.scene.collection.children.link(col)

            # Create empties and add them to pairs[]
            pairs = []
            for bone_obj in context.selected_pose_bones_from_active_object:
                empty = bpy.data.objects.new(
                    self.bone_prefix + bone_obj.name, None)
                if self.use_additive:
                    empty.empty_display_type = 'CUBE'
                else:
                    empty.empty_display_type = 'SPHERE'
                empty.empty_display_size = 0.1

                col.objects.link(empty)

                empty_add = None
                if self.use_additive:
                    empty_add = bpy.data.objects.new(
                        'additive_' + bone_obj.name, None)
                    empty_add.empty_display_type = 'SPHERE'
                    empty_add.empty_display_size = 0.1

                    col.objects.link(empty_add)

                    empty_add.parent = empty

                empty.parent = context.object
                empty.parent_bone = bone_obj.name
                empty.parent_type = 'BONE'

                empty.location.y -= 1.0
                if self.axis == 'X':
                    empty.location.x = context.active_pose_bone.bone.length * self.length
                if self.axis == 'Y':
                    empty.location.y = context.active_pose_bone.bone.length * \
                        self.length - context.active_pose_bone.bone.length
                if self.axis == 'Z':
                    empty.location.z = context.active_pose_bone.bone.length * self.length

                pairs.append((bone_obj, empty, empty_add))

            # Bake
            if self.bake:
                con_temp = context.copy()
                con_temp['selected_editable_objects'] = [p[1] for p in pairs]

                if self.use_scene:
                    self.frame_start = context.scene.frame_start
                    self.frame_end = context.scene.frame_end

                # Bake to object space by evaluating then applying matrices
                if self.use_custom_space:
                    self.bake_to_custom_obj_space(context, con_temp)

                # Bake to world space by using regular bake operator
                else:
                    bpy.ops.nla.bake(
                        con_temp,
                        frame_start=self.frame_start,
                        frame_end=self.frame_end,
                        clear_parents=True,
                        visual_keying=True,
                        bake_types={'OBJECT'}
                    )

            for bone_obj, empty, empty_add in pairs:
                con = bone_obj.constraints.new(self.constraint)

                if empty_add is not None:
                    con.target = empty_add
                else:
                    con.target = empty

                if self.axis == 'X':
                    if self.length < 0.0:
                        con.track_axis = 'TRACK_NEGATIVE_X'
                    else:
                        con.track_axis = 'TRACK_X'

                if self.axis == 'Y':
                    if self.length < 0.0:
                        con.track_axis = 'TRACK_NEGATIVE_Y'
                    else:
                        con.track_axis = 'TRACK_Y'

                if self.axis == 'Z':
                    if self.length < 0.0:
                        con.track_axis = 'TRACK_NEGATIVE_Z'
                    else:
                        con.track_axis = 'TRACK_Z'

            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            bpy.context.window.cursor_modal_set('DEFAULT')
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}

    def invoke(self, context, event):
        self.start_x = event.mouse_x
        self.start_y = event.mouse_y
        return context.window_manager.invoke_props_dialog(self)

    def draw(self, context):
        layout = self.layout
        row = layout.row(align=True)
        row.prop(self, 'bake')
        row.prop(self, 'use_scene')
        if self.bake and not self.use_scene:
            row = layout.row(align=True)
            row.prop(self, 'frame_start')
            row.prop(self, 'frame_end')
        layout.prop(self, 'subcol_name')
        layout.prop(self, 'bone_prefix')
        layout.prop(self, 'constraint')
        layout.prop(self, 'use_additive')

        if context.mode == 'POSE' and self.constraint == 'DAMPED_TRACK':
            layout.prop(self, 'use_custom_axis')

        layout.prop(self, 'use_custom_space')
        if self.use_custom_space:
            layout.prop_search(self, 'custom_space', context.scene, "objects")
            ob = context.scene.objects.get(self.custom_space)
            if ob is not None and ob.type == 'ARMATURE':
                layout.prop_search(self, 'custom_space_bone', ob.data, "bones")

    def bake_to_custom_obj_space(self, context, con_temp):
        space_parent = context.scene.objects.get(self.custom_space)
        if space_parent is None:
            self.report(
                {'ERROR'}, f"Object {self.custom_space} not found in Scene!")
            return {'CANCELLED'}
        elif space_parent in con_temp['selected_editable_objects']:
            self.report({'ERROR'}, f"Object {self.custom_space} "
                                   f"selected as parent, but was created as a bake target!")
            return {'CANCELLED'}
        # Build a cache of worldspace matrices for the empties, prebuild list for perf
        world_cache = {obj: [None] * (self.frame_end - self.frame_start + 1) for obj in
                       con_temp['selected_editable_objects']}

        # Update matrices after parent
        context.view_layer.depsgraph.update()

        # Get world matrices on each frame
        for i in range(self.frame_end - self.frame_start + 1):
            context.scene.frame_set(self.frame_start + i)
            for obj in world_cache.keys():
                # Is copy needed? Probably. Is it slow? Definitely.
                world_cache[obj][i] = obj.matrix_world.copy()

        # Parent objs to camera
        for obj in world_cache.keys():
            obj.parent_type = 'OBJECT'
            obj.parent = space_parent
            if self.custom_space_bone != "" and obj.parent.type == 'ARMATURE':
                obj.parent_type = 'BONE'
                obj.parent_bone = self.custom_space_bone

        # Update view layer after changing parents again, just in case.
        context.view_layer.depsgraph.update()

        # Brute force section - adding keyframes is O(n^2) with number of keyframes to add
        # If it's too slow then extract the loc rot scale
        #   from the matrix and add fcurve points directly (foreach_set etc.)
        for i in range(self.frame_end - self.frame_start + 1):
            f = self.frame_start + i
            context.scene.frame_set(f)
            for obj in world_cache.keys():
                obj.matrix_world = world_cache[obj][i]
                obj.keyframe_insert("location", index=-1, frame=f)
                obj.keyframe_insert("scale", index=-1, frame=f)
                obj.keyframe_insert("rotation_euler", index=-1, frame=f)

    def execute(self, context):

        if context.mode == 'POSE' and self.constraint == 'DAMPED_TRACK' and self.use_custom_axis:

            if context.area.type == 'VIEW_3D':
                if bpy.context.active_pose_bone != None:
                    # the arguments we pass the the callback
                    args = (self, context)
                    # Add the region OpenGL drawing callback
                    # draw in view space with 'POST_VIEW' and 'PRE_VIEW'
                    self._handle = bpy.types.SpaceView3D.draw_handler_add(
                        draw_callback_px, args, 'WINDOW', 'POST_VIEW')

                    self.emp_loc = []
                    self.length = 1.0
                    self.axis = 'Y'

                    bo = context.active_pose_bone

                    self.x_line = [bo.head + bo.x_axis *
                                   10, bo.head - bo.x_axis * 10]
                    self.y_line = [bo.head + bo.y_axis *
                                   10, bo.head - bo.y_axis * 10]
                    self.z_line = [bo.head + bo.z_axis *
                                   10, bo.head - bo.z_axis * 10]

                    bpy.context.window.cursor_modal_set('NONE')
                    context.window_manager.modal_handler_add(self)
                    return {'RUNNING_MODAL'}
                else:
                    self.report(
                        {'WARNING'}, "No active bone selected")
                    return {'CANCELLED'}
            else:
                self.report(
                    {'WARNING'}, "View3D not found, cannot run operator")
                return {'CANCELLED'}

        else:

            col = bpy.data.collections.new(self.subcol_name)
            parent = context.collection

            if parent and not is_a_library(parent):
                context.collection.children.link(col)
            else:
                context.scene.collection.children.link(col)

            is_posemode = context.mode == 'POSE'

            iter_target = context.selected_pose_bones_from_active_object if \
                is_posemode else context.selected_editable_objects

            # Create empties and add them to pairs[]
            pairs = []
            for bone_obj in iter_target:
                empty = bpy.data.objects.new(
                    self.bone_prefix + bone_obj.name, None)
                empty.empty_display_type = 'CUBE'
                empty.empty_display_size = 0.1

                col.objects.link(empty)

                empty_add = None
                if self.use_additive:
                    empty_add = bpy.data.objects.new(
                        'additive_' + bone_obj.name, None)
                    empty_add.empty_display_type = 'SPHERE'
                    empty_add.empty_display_size = 0.1

                    col.objects.link(empty_add)

                    empty_add.parent = empty

                empty.parent = context.object
                empty.parent_bone = bone_obj.name
                empty.parent_type = 'BONE' if is_posemode else 'OBJECT'

                if not self.constraint == 'DAMPED_TRACK' and is_posemode:
                    empty.location.y -= bone_obj.bone.length

                pairs.append((bone_obj, empty, empty_add))

            # Bake
            if self.bake:
                con_temp = context.copy()
                con_temp['selected_editable_objects'] = [p[1] for p in pairs]

                if self.use_scene:
                    self.frame_start = context.scene.frame_start
                    self.frame_end = context.scene.frame_end

                # Bake to object space by evaluating then applying matrices
                if self.use_custom_space:
                    self.bake_to_custom_obj_space(context, con_temp)

                # Bake to world space by using regular bake operator
                else:
                    bpy.ops.nla.bake(
                        con_temp,
                        frame_start=self.frame_start,
                        frame_end=self.frame_end,
                        clear_parents=True,
                        visual_keying=True,
                        bake_types={'OBJECT'}
                    )

            if self.constraint != 'NONE':
                for bone_obj, empty, empty_add in pairs:
                    con = bone_obj.constraints.new(self.constraint)

                    if empty_add is not None:
                        con.target = empty_add
                    else:
                        con.target = empty

        return {'FINISHED'}


def pose_menu_draw(self, _ctx):
    layout = self.layout
    layout.separator()
    layout.operator(ARM_OT_BakeBonesToEmpties.bl_idname)


class ARM_OT_ConvertEulerOrder(Operator):
    bl_idname = 'object.convert_euler_order'
    bl_label = 'Convert Euler Rotation Order'
    bl_options = {'UNDO'}

    new_order: EnumProperty(
        name="Euler Order",
        description="New Euler order to convert animation into",
        items=[('XYZ', "XYZ", "Convert to XYZ order"),
               ('XZY', "XZY", "Convert to XZY order"),
               ('YXZ', "YXZ", "Convert to YXZ order"),
               ('YZX', "YZX", "Convert to YZX order"),
               ('ZXY', "ZXY", "Convert to ZXY order"),
               ('ZYX', "ZYX", "Convert to ZYX order")],
        default='XYZ'
    )

    @classmethod
    def poll(cls, context):
        return context.object is not None and context.mode in {'OBJECT', 'POSE'}

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self)

    def execute(self, context):
        ob = bpy.context.active_object

        old_order = ob.rotation_mode

        if ob.animation_data is not None and ob.animation_data.action is not None:

            xa, ya, za = None, None, None
            frames = []
            for fc in ob.animation_data.action.fcurves:
                if 'rotation_euler' in fc.data_path:
                    if fc.array_index == 0:
                        xa = fc
                    if fc.array_index == 1:
                        ya = fc
                    if fc.array_index == 2:
                        za = fc

                    frames += [f.co[0] for f in fc.keyframe_points]

            frames = list(set(frames))

            # Found at least 1 axis to convert keyframes
            if xa != None or ya != None or za != None:
                for f in frames:
                    rots = [0, 0, 0]

                    xk, yk, zk = None, None, None

                    if xa is None:
                        rots[0] = ob.rotation_euler[0]
                    else:
                        for k in xa.keyframe_points:
                            if k.co[0] == f:
                                rots[0] = k.co[1]
                                xk = k
                                break

                    if ya is None:
                        rots[1] = ob.rotation_euler[1]
                    else:
                        for k in ya.keyframe_points:
                            if k.co[0] == f:
                                rots[1] = k.co[1]
                                yk = k
                                break

                    if za is None:
                        rots[2] = ob.rotation_euler[2]
                    else:
                        for k in za.keyframe_points:
                            if k.co[0] == f:
                                rots[2] = k.co[1]
                                zk = k
                                break

                    euler = Euler(rots, old_order)
                    quat = euler.to_quaternion()
                    new_euler = quat.to_euler(self.new_order)

                    if xk:
                        xk.co[1] = new_euler[0]

                    if yk:
                        yk.co[1] = new_euler[1]

                    if zk:
                        zk.co[1] = new_euler[2]

        quat = ob.rotation_euler.to_quaternion()
        ob.rotation_mode = self.new_order
        ob.rotation_euler = quat.to_euler(self.new_order)

        return {'FINISHED'}


_classes = [
    ARM_OT_BakeBonesToEmpties,
    ARM_OT_ConvertEulerOrder,
    NLA_OT_smart_bake,
]

_register, _unregister = bpy.utils.register_classes_factory(_classes)


def register():
    _register()
    bpy.types.VIEW3D_MT_object_animation.append(pose_menu_draw)


def unregister():
    _unregister()
    bpy.types.VIEW3D_MT_object_animation.remove(pose_menu_draw)
