import bpy
from bpy.props import *
from bpy.types import Panel, Operator


class GRAPH_OT_quick_noise(Operator):
    bl_idname = 'graph.quick_noise'
    bl_label = "Quick Noise"
    bl_options = {'REGISTER', 'UNDO'}

    frame_start: FloatProperty(
        name="Start",
        default=1
    )

    blend_in: FloatProperty(
        name="Blend In",
        default=1.0,
        min=0.0
    )

    frame_end: FloatProperty(
        name="End",
        default=100
    )

    blend_out: FloatProperty(
        name="Blend Out",
        default=-1.0,
        max=0.0
    )

    scale: FloatProperty(
        name="Scale",
        default=1.0
    )

    amplitude: FloatProperty(
        name="Amplitude",
        default=1.0
    )

    seed: IntProperty(
        name="Seed",
        default=0
    )

    reset: BoolProperty(
        name="Reset Frame Range",
        default=False
    )

    @classmethod
    def poll(cls, ctx: bpy.types.Context):
        return \
            ctx.space_data.type == 'GRAPH_EDITOR' and \
            len(ctx.selected_editable_fcurves) > 0

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        col = row.column(align=True)
        col.prop(self, "reset", emboss=True, toggle=1)
        row = col.row(align=True)
        row.prop(self, "frame_start")
        row.prop(self, "frame_end")
        row = col.row(align=True)
        row.prop(self, "blend_in")
        row.prop(self, "blend_out")

        layout.separator()

        col = layout.column()
        col.prop(self, "scale")
        col.prop(self, "amplitude")
        col.prop(self, "seed")

    def invoke(self, ctx, event):
        if self.reset:
            self.frame_start = ctx.scene.frame_start
            self.frame_end = ctx.scene.frame_end
        for curve in ctx.selected_editable_fcurves:
            mod = curve.modifiers.new('NOISE')
            mod.use_restricted_range = True
        self.execute(ctx)
        return ctx.window_manager.invoke_props_popup(self, event)

    def cancel(self, ctx):
        # Cleanup
        for c in ctx.selected_editable_fcurves:
            c.modifiers.remove(c.modifiers[-1])

    def execute(self, ctx: bpy.types.Context):
        if self.reset:
            self.reset = False
            self.frame_start = ctx.scene.frame_start
            self.frame_end = ctx.scene.frame_end
            self.blend_in = 1.0
            self.blend_out = -1.0
        n = len(ctx.selected_editable_fcurves)
        for i, c in enumerate(ctx.selected_editable_fcurves):
            mod = c.modifiers[-1]
            mod.phase = float((i * n) + self.seed)
            mod.strength = self.amplitude
            mod.scale = self.scale
            mod.frame_start = self.frame_start
            mod.frame_end = self.frame_end
            mod.blend_in = self.blend_in
            mod.blend_out = -self.blend_out
        return {'FINISHED'}


class GRAPH_OT_pop_noise(Operator):
    bl_idname = 'graph.pop_noise'
    bl_label = "Pop Noise"
    bl_options = {'REGISTER', 'UNDO'}

    pop_all: BoolProperty(
        "Pop All",
        default=False
    )

    @classmethod
    def poll(cls, ctx: bpy.types.Context):
        return \
            ctx.space_data.type == 'GRAPH_EDITOR' and \
            len(ctx.selected_editable_fcurves) > 0

    def execute(self, ctx):
        for c in ctx.selected_editable_fcurves:
            if self.pop_all:
                mods = (m for m in c.modifiers)
                for m in mods:
                    if m.type == 'NOISE':
                        c.modifiers.remove(m)
            else:
                mod = c.modifiers[-1]
                if mod.type == 'NOISE':
                    c.modifiers.remove(mod)
            c.update()
        return {'FINISHED'}


class GRAPH_PT_QuickFX(Panel):
    bl_label = "Quick FX"
    bl_space_type = 'GRAPH_EDITOR'
    bl_region_type = 'UI'
    bl_category = 'Tool'

    def draw(self, context):
        layout = self.layout
        op = layout.operator("graph.quick_noise")
        op.reset = True
        op = layout.operator("graph.pop_noise", text="Pop Noise",
                             text_ctxt="Remove last noise modifier from selected fcurves")
        op.pop_all = False
        op = layout.operator("graph.pop_noise", text="Clear Noise",
                             text_ctxt="Remove all nosie modifiers from selected fcurves")
        op.pop_all = True
        layout.separator()
        op = layout.operator("graph.quick_noise", text="Preset: Camera Shake")
        op.frame_start = context.scene.frame_start
        op.frame_end = op.frame_start + 5
        op.amplitude = 0.05


def key_menu_draw(self, _ctx):
    layout = self.layout
    layout.separator()
    op = layout.operator(GRAPH_OT_quick_noise.bl_idname)
    op.reset = True
    op = layout.operator("graph.pop_noise", text="Pop Noise",
                         text_ctxt="Remove last noise modifier from selected fcurves")
    op.pop_all = False
    op = layout.operator("graph.pop_noise", text="Clear Noise",
                         text_ctxt="Remove all nosie modifiers from selected fcurves")
    op.pop_all = True


_classes = [
    GRAPH_OT_quick_noise,
    GRAPH_OT_pop_noise,
    GRAPH_PT_QuickFX
]

_register, _unregister = bpy.utils.register_classes_factory(_classes)


def register():
    _register()
    bpy.types.GRAPH_MT_key.append(key_menu_draw)


def unregister():
    _unregister()
    bpy.types.GRAPH_MT_key.remove(key_menu_draw)
